/***************************************************************************\
  "parser.h" - Parser Module (Phase C)
  Revision 14; 10 Nov 2002 - 15 Jan 2003
  Copyright (C) 2002-2003  Kevin P. Albrecht
\***************************************************************************/

#ifndef PARSER_H
#define PARSER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "global.h"
#include "error.h"
#include "icm.h"
#include "k_string.h"
#include "global_conv.h"
#include "table.h"
#include "scope.h"
#include "token.h"
#include "tempgen.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Internal //
Token *mToken;    // current token
TokenCode mCode;  // code of current token
bool bEndOfBlock; // tells the parser_parse function to ignore the lack
                  //  of an EOL because a block just ended.
Symbol *mSymNL, *mSymCINPUT, *mSymMAIN;
Routine *mCurrRout;

// External //
void parser_init ();
void parser_deinit ();
void parser_parse ();

// Internal - Special Parse //
void mParseStatement ();

// Internal - Parse Statements //
void mParseLabel ();
void mParseGoto ();
void mParseDefinition ();
void mParseConstDefinition ();
void mParseIdentifierStatement ();
void mParsePrint ();
void mParsePush ();
void mParsePop ();
void mParseExit ();

// Internal - Parse Blocks //
void mParseBlock (int anInitialIndent);
void mParseRoutine ();
void mParseReturn ();
void mParseCond ();
void mParseLoop ();

// Internal - Parse Expressions //
Symbol *mParseExpression ();       // =, <>, <, >, <=, >=
Symbol *mParseSimpleExpression (); // +u, -u, +, -, or
Symbol *mParseTerm ();             // *, /, and
Symbol *mParseFactor ();           // identifier, literal, paren expression

// Internal - Special //
void mSkipEOL ();
void mEnsureSkipEOL ();
bool mSync (TokenCode aCode, char *aError);
void mSeekToEOL ();
void mGetToken ();

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// parser_init
//// Initialize parser module
void parser_init ()
{
  // Blank right now //
}

//// parser_deinit
//// Deinitialize parser module
void parser_deinit ()
{
  // Blank right now //
}

//// parser_parse
void parser_parse ()
{
  bEndOfBlock = false;

  mSymNL = scope_global_enter ("nl");
  sym_set (mSymNL, repIdent, tStr, true/*const*/, k_clone("\n"));

  mSymCINPUT = scope_global_enter ("cinput");
  sym_set (mSymCINPUT, repIdent, tStr, true/*const*/, k_clone("<cinput>"));

  mSymMAIN = NULL;
  mCurrRout = NULL;

  // put a 'nop' instruction as a placeholder for when //
  //  the 'main' routine is encoutered                 //                                           //
  icm_put (statement_new (tacNOP, NULL, NULL, NULL));

  // put a 'stop' instruction to stop the program //
  icm_put (statement_new (tacSTOP, NULL, NULL, NULL));

  /////////////
  // PARSING //
  /////////////

  mToken = tok_new ();      //
  mToken->code = codeStart; // Passed to lex_get at the start of the file
  mGetToken();              //
  mSkipEOL();

  do {
    // Insert a line number marker //
    icm_putline ();

    // Parse a Statement //
    mParseStatement ();

    // Handles empty files //
    if (mCode == codeEOF) break;

    // Unless a block just ended, check for EOL //
    if (bEndOfBlock == true)
      bEndOfBlock = false;
    else
      mEnsureSkipEOL();

  } while (mCode != codeEOF);

  if (mSymMAIN == NULL)
  {
    err (true, EF_MAIN_NOT_DEF, NULL);
    erre (EFF_RECOVER);
  }
  else if ( ((Routine*)(mSymMAIN->value))->returnType != tVoid )
  {
    err (true, EF_MAIN_NOT_VOID, gvsTypeStrings[((Routine*)(mSymMAIN->value))->returnType]);
    erre (EFF_RECOVER);
  }
}

///////////////////
// Special Parse //
///////////////////

//// mParseStatement
void mParseStatement ()
{
  switch (mCode)
  {
    case (codePOUNDSIGN):
      mParseLabel ();
      break;
    case (codeGOTO):
      mParseGoto ();
      break;  
    case (codeDEF):
      mParseDefinition ();
      break;
    case (codeCONST):
      mParseConstDefinition ();
      break;
    case (codeIdentifier):
      mParseIdentifierStatement ();
      break;
    case (codePRINT):
    case (codePRINTN):
      mParsePrint ();
      break;
    case (codePUSH):
      mParsePush ();
      break;
    case (codePOP):
      mParsePop ();
      break;
    case (codeEXIT):
      mParseExit ();
      break;
    case (codeROUT):
      mParseRoutine ();
      break;
    case (codeRETURN):
      mParseReturn ();
      break;
    case (codeCOND):
      mParseCond ();
      break;
    case (codeLOOP):
      mParseLoop ();
      break;
    case (codeEOL):
    case (codeEOF):
      break;
    default:
      err (true, EF_INVALID_PARSE_POINT, gvsTokenCodeStrings[mCode]);
      erre (EFF_TERMINATE);
      break;
  }
}

///////////////////////
// Statement Parsing //
///////////////////////

//// mParseLabel
void mParseLabel ()
{
  Symbol *symbolA;

  mGetToken (); // now: identifier

  // Search for the label's identifier and enter it //
  symbolA = scope_global_search  (mToken->lexeme);
  if (symbolA) // label must have been created by mParseGoto
  {
    if (symbolA->type == tLabel)
    {
      symbolA->value = k_new_n(gxCurr);
      symbolA->constant = true; // means that label has been defined
    }
    else
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbolA->type]);
      erre (EFF_RECOVER);
    }
  }
  else if (!symbolA) // create label
  {
    symbolA = scope_global_enter (mToken->lexeme);
    sym_set (symbolA, repLiteral, tLabel, false, k_new_n(gxCurr));
    symbolA->constant = true; // means that label has been defined
  }

  mGetToken (); // now: EOL or statement

  // The label is directly followed by a statement //
  if (mCode != codeEOL)
  {
    mParseStatement ();
  }
}

//// mParseGoto
void mParseGoto ()
{
  Symbol *symbolA;

  mGetToken (); // now: identifier

  // Search for the label's identifier and enter it //
  symbolA = scope_global_search  (mToken->lexeme);
  if (symbolA) // label must have been created already
  {
    if (symbolA->type == tLabel)
    {
      icm_put (statement_new (tacJUMP, symbolA, NULL, NULL));
    }
    else
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbolA->type]);
      erre (EFF_RECOVER);
    }
  }
  else if (!symbolA) // create label
  {
    symbolA = scope_global_enter (mToken->lexeme);
    sym_set (symbolA, repLiteral, tLabel, false, NULL);
    symbolA->constant = false; // means that label is not yet defined;

    icm_put (statement_new (tacJUMP, symbolA, NULL, NULL));
  }

  mGetToken (); // now: EOL
}

//// mParseDefinition
void mParseDefinition ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  char *lexemeA;
  Type typeA;

  while (true)
  {
    mGetToken (); // now: identifier

    lexemeA = k_clone (mToken->lexeme);

    mGetToken (); // now: ':'
    if (!mSync (codeCOLON, EF_MISSING_TYPE_DEF_OP)) return;

    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) typeA = tInt;
    else if (mCode == codeRAT) typeA = tRat;
    else if (mCode == codeSTR) typeA = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      erre (EFF_RECOVER);
      typeA = tInt; // set as int for recovery
    }

    mGetToken (); // now: '<-' or EOL

    // definition is also an assignment //
    if (mCode == codeARROW)
    {
      mGetToken (); // now expression
      symbolB = mParseExpression ();

      if (!((((typeA == tInt) || (typeA == tRat)) &&
             ((symbolB->type == tInt) || (symbolB->type == tRat))) ||
            (((typeA == tStr) && (symbolB->type == tStr)))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }
    }
    // definition is not an assignment //
    else
    {
      symbolB = NULL; // set to NULL so that error won't occur later
    }

    // Search for the variable's identifier and enter it //
    symbolA = scope_local_search_only  (lexemeA);
    if (symbolA)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DEFINED, lexemeA);
      erre (EFF_RECOVER);
    }
    else if (!symbolA)
    {
      // The symbol is created //
      symbolA = scope_local_enter (lexemeA);
      sym_set (symbolA, repIdent, typeA, false, NULL);
    }

    // Insert intermediate code if also assignment //
    if (symbolB)
    {
      tempgen_check (symbolB);
      icm_put (statement_new (tacSET, symbolA, symbolB, NULL));
    }

    k_del (lexemeA);

    if (mCode != codeCOMMA) // there is another variable defined
    {
      break;
    }
  }
}

//// mParseConstDefinition
void mParseConstDefinition ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  char *lexemeA;
  Type typeA;

  while (true)
  {
    mGetToken (); // now: identifier

    lexemeA = k_clone (mToken->lexeme);

    mGetToken (); // now: ':'
    if (!mSync (codeCOLON, EF_MISSING_TYPE_DEF_OP)) return;

    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) typeA = tInt;
    else if (mCode == codeRAT) typeA = tRat;
    else if (mCode == codeSTR) typeA = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      erre (EFF_RECOVER);
      typeA = tInt; // set as int for recovery
    }

    mGetToken (); // now: '<-'
    if (!mSync (codeARROW, EF_MISSING_EXPRESSION)) return;


    mGetToken (); // now expression
    symbolB = mParseExpression ();

    if (!((((typeA == tInt) || (typeA == tRat)) &&
           ((symbolB->type == tInt) || (symbolB->type == tRat))) ||
          (((typeA == tStr) && (symbolB->type == tStr)))
         )
       )
    {
      err (true, EF_TYPE_INCOMPATIBLE, NULL);
      erre (EFF_RECOVER);
    }

    // Search for the variable's identifier and enter it //
    symbolA = scope_local_search_only  (lexemeA);
    if (symbolA)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DEFINED, lexemeA);
      erre (EFF_RECOVER);
    }
    else if (!symbolA)
    {
      // The symbol is created //
      symbolA = scope_local_enter (lexemeA);
      sym_set (symbolA, repIdent, typeA, true/*const*/, NULL);
    }

    // Insert intermediate code //
    tempgen_check (symbolB);
    icm_put (statement_new (tacSET, symbolA, symbolB, NULL));

    k_del (lexemeA);

    if (mCode != codeCOMMA) // there is another const defined
    {
      break;
    }
  }
}

//// mParseIdentifierStatement
void mParseIdentifierStatement ()
{
  Symbol *symbolA;
  Symbol *symbolB;

  Statement *temp;

  // Search for the identifier and enter it if necessary //
  symbolA = scope_local_search (mToken->lexeme);
  if (!symbolA)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);
    erre (EFF_RECOVER);

    // Enter the symbol as int for recovery //
    symbolA = scope_local_enter (mToken->lexeme);
    sym_set (symbolA, repIdent, tInt, false, k_new_n(0));
  }

  // Statement is a Routine call //
  if (symbolA->type == tRoutine)
  {
    temp = statement_new (tacCALL, symbolA, NULL, NULL);
    icm_put (temp);
    mGetToken (); // now EOL
  }

  // Statement is an Assignment //
  else
  {
    // Ensure no assignment to constant //
    if (symbolA->constant == true)
    {
      err (true, EF_CONST_ASSIGNMENT, mToken->lexeme);
      erre (EFF_RECOVER);
    }

    // :
    mGetToken ();
    if (!mSync (codeARROW, EF_MISSING_ASSIGNMENT_OP)) return;

    // Expression //
    mGetToken ();
    symbolB = mParseExpression ();

    if (!((((symbolA->type == tInt) || (symbolA->type == tRat)) &&
           ((symbolB->type == tInt) || (symbolB->type == tRat))) ||
          (((symbolA->type == tStr) && (symbolB->type == tStr)))
         )
       )
    {
      err (true, EF_TYPE_INCOMPATIBLE, NULL);
      erre (EFF_RECOVER);
    }

    tempgen_check (symbolB);
    icm_put (statement_new (tacSET, symbolA, symbolB, NULL));
  }
}

//// mParsePrint
void mParsePrint ()
{
  bool newline;
  Symbol *symbol;

  // check to see if new line should be printed //
  if (mCode == codePRINTN) newline = true;
  else                     newline = false;

  mGetToken (); // now first token of exp or EOL
  
  if (mCode != codeEOL)
  {
    symbol = mParseExpression ();

    tempgen_check (symbol);
    icm_put (statement_new (tacCOUT, symbol, NULL, NULL));
  }
    
  if (newline)
  {
    icm_put (statement_new (tacCOUT, mSymNL, NULL, NULL));
  }
}

void mParsePush ()
{
  Symbol *symbol;

  mGetToken (); // now first token of exp

  symbol = mParseExpression ();

  tempgen_check (symbol);
  icm_put (statement_new (tacPUSH, symbol, NULL, NULL));
}

void mParsePop ()
{
  Symbol *symbol;

  mGetToken (); // now target name

  // Search for the target variable's identifier and enter it if necessary //
  symbol = scope_local_search (mToken->lexeme);
  if (!symbol)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);
    erre (EFF_RECOVER);

    // Enter the symbol as int for recovery //
    symbol = scope_local_enter (mToken->lexeme);
    sym_set (symbol, repIdent, tInt, false, k_new_n(0));
  }

  // Ensure no assignment to constant //
  if (symbol->constant == true)
  {
    err (true, EF_CONST_ASSIGNMENT, mToken->lexeme);
    erre (EFF_RECOVER);
  }

  icm_put (statement_new (tacPOP, symbol, NULL, NULL));

  mGetToken (); // now: EOL
}

// mParseExit //
void mParseExit ()
{
  icm_put (statement_new (tacSTOP, NULL, NULL, NULL));
  mGetToken (); // now: EOL
}

///////////////////
// Block Parsing //
///////////////////

//// mParseBlock
//// Created: 0142 11/24/02
void mParseBlock (int anInitialIndent)
{
  int i = 0;

  while ((mCode != codeEOF) && (gnIndent > anInitialIndent))
  {
    // Insert a line number marker //
    icm_putline ();

    // Parse a Statement //
    mParseStatement ();

    // Unless a block just ended, check for EOL //
    if (bEndOfBlock == true) bEndOfBlock = false;
    else                     mEnsureSkipEOL();

    i++;
  }    

  if (i == 0)
  {
    err (true, EF_EMPTY_BLOCK, NULL);
    erre (EFF_RECOVER);
  }

  bEndOfBlock = true;
}

//// mParseRoutine
void mParseRoutine ()
{
  Symbol *symRoutine;
  Symbol *symTemp;
  int nInitialIndent;
  Type returnType;

  nInitialIndent = gnIndent;
  mGetToken (); // now: identifier

  // Search for the routine's identifier and enter it //
  symRoutine = scope_global_search  (mToken->lexeme);
  if (!symRoutine)
  {
    symRoutine = scope_global_enter (mToken->lexeme);

    // If this is the 'main' routine, insert the code to execute it //
    if (strcmp (mToken->lexeme, "main") == 0)
    {
      mSymMAIN = symRoutine;

      // delete the 'nop' and set the first instruction //
      statement_del (gvProgram[0]);
      gvProgram[0] = statement_new (tacCALL, mSymMAIN, NULL, NULL);
    }
  }
  else
  {
    err (true, EF_ROUT_REDEF, mToken->lexeme);
    erre (EFF_TERMINATE);
  }

  mGetToken (); // now: ':' or EOL

  // Routine has a non-void type //
  if (mCode == codeCOLON)
  {
    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) returnType = tInt;
    else if (mCode == codeRAT) returnType = tRat;
    else if (mCode == codeSTR) returnType = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      erre (EFF_RECOVER);
      returnType = tInt; // set as int for recovery
    }

    sym_set (symRoutine, repLiteral, tRoutine, false, routine_new(returnType, gxCurr));

    mGetToken (); // now: EOL
  }
  // Routine has a void type //
  else
  {
    sym_set (symRoutine, repLiteral, tRoutine, false, routine_new(tVoid, gxCurr));
  }

  mEnsureSkipEOL(); // now: first token of block

  // Parse a block of statements //
  mCurrRout = (Routine*)(symRoutine->value);
  mParseBlock (nInitialIndent);

  // Make sure that routines with non-void return types put //
  //  a value on the stack.                                 //
  if (mCurrRout->returnType != tVoid)
  {
    if (mCurrRout->returnedFrom == false)
    {
      err (true, EF_RETURN_NEVER_MADE, symRoutine->key);
      erre (EFF_RECOVER);
    }

    symTemp = sym_new ("routine_error_exit");
    sym_set (symTemp, repLiteral, ((Routine*)(symRoutine->value))->returnType, true, NULL);
    icm_put (statement_new (tacPUSH, symTemp, NULL, NULL));
  }

  // Put a 'ret' in the icode in case one was never placed in //
  //  the body of the routine.                                //
  icm_put (statement_new (tacRET, NULL, NULL, NULL));

  // Let the rest of the parser know that it is not in a routine //
  mCurrRout = NULL;
}

//// mParseReturn
void mParseReturn ()
{
  Symbol *symbol;

  if (mCurrRout != NULL)
  {
    mGetToken (); // now first token of exp or EOL

    // Routine does not have type void; push value to stack //
    if (mCurrRout->returnType != tVoid)
    {
      // push return value to stack //
      symbol = mParseExpression ();

      mCurrRout->returnedFrom = true;

      // expression type does not match routine return type //
      if (symbol->type != mCurrRout->returnType)
      {
        err (true, EF_NON_MATCHING_RET_TYPE, gvsTypeStrings[mCurrRout->returnType]);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbol);
      icm_put (statement_new (tacPUSH, symbol, NULL, NULL));
    }
    else
    {
      if (mCode != codeEOL)
      {
        err (true, EF_NO_RETURN_FOR_VOID_ROUT, NULL);
        erre (EFF_RECOVER);
        mSeekToEOL ();
      }
    }

    icm_put (statement_new (tacRET, NULL, NULL, NULL));
  }
  else
  {
    err (true, EF_RET_OUTSIDE_OF_ROUT, NULL);
    erre (EFF_RECOVER);
  }
}

//// mParseCond
void mParseCond ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Symbol *symbolC;
  Statement *statementCOND;
  int nInitialIndent;

  nInitialIndent = gnIndent;

  mGetToken (); // now first token of expression
  symbolA = mParseExpression ();

  tempgen_check (symbolA);
  statementCOND = statement_new (tacCOND, symbolA, NULL, NULL);
  icm_put (statementCOND);

  if ((symbolA->type != tInt) && (symbolA->type != tRat))
  {
    err (true, EF_TYPE_INCOMPATIBLE, "\'cond\' accepts \'int\' or \'rat\'");
    erre (EFF_RECOVER);
  }

  mEnsureSkipEOL();

  symbolB = sym_new ("cond_true_pos");
  sym_set (symbolB, repLiteral, tLabel, false, k_new_n (gxCurr));

  // parse a block of statements //
  mParseBlock (nInitialIndent);

  symbolC = sym_new ("cond_false_pos");
  sym_set (symbolC, repLiteral, tLabel, false, k_new_n (gxCurr));

  statementCOND->b = symbolB;
  statementCOND->c = symbolC;
}

//// mParseLoop
//// Created: 0148 11/24/02
void mParseLoop ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Symbol *symbolC;
  Symbol *symbolStart;
  Statement *statementCOND;

  int nInitialIndent;

  nInitialIndent = gnIndent;

  // Set the start of the loop symbol //
  symbolStart = sym_new ("loop_start_pos");
  sym_set (symbolStart, repLiteral, tLabel, true, k_new_n(gxCurr));

  mGetToken (); // now first token of expression
  symbolA = mParseExpression ();

  tempgen_check (symbolA);
  statementCOND = statement_new (tacCOND, symbolA, NULL, NULL);
  icm_put (statementCOND);

  if ((symbolA->type != tInt) && (symbolA->type != tRat))
  {
    err (true, EF_TYPE_INCOMPATIBLE, "\'cond\' accepts \'int\' or \'rat\'");
    erre (EFF_RECOVER);
  }

  mEnsureSkipEOL();

  // Set the true position symbol //
  symbolB = sym_new ("cond_true_pos");
  sym_set (symbolB, repLiteral, tLabel, false, k_new_n (gxCurr));

  // parse a block of statements //
  mParseBlock (nInitialIndent);

  icm_put (statement_new (tacJUMP, symbolStart, NULL, NULL));

  // Set the false position symbol //
  symbolC = sym_new ("cond_false_pos");
  sym_set (symbolC, repLiteral, tLabel, false, k_new_n (gxCurr));

  statementCOND->b = symbolB;
  statementCOND->c = symbolC;
}

////////////////////////
// Expression Parsing //
////////////////////////

//// mParseExpression
//// Duties: relational ops
Symbol *mParseExpression ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Symbol *symbolC;
  TAC tac;

  // Parse a simple expression //
  symbolB = mParseSimpleExpression ();

  // If there is a relational operator, parse another simple expression //
  while ((mCode == codeEQUALS) || (mCode == codeNOTEQUAL) ||
         (mCode == codeLT) || (mCode == codeGT) ||
         (mCode == codeLTEQUAL) || (mCode == codeGTEQUAL)
        )
  {
    if (mCode == codeEQUALS) tac = tacEQ;
    else if (mCode == codeNOTEQUAL) tac = tacNEQ;
    else if (mCode == codeLT) tac = tacLT;
    else if (mCode == codeGT) tac = tacGT;
    else if (mCode == codeLTEQUAL) tac = tacLTEQ;
    else if (mCode == codeGTEQUAL) tac = tacGTEQ;

    // ensure that strings are not used with anything but '=' and '<>' //
    if ((tac != tacEQ) && (tac != tacNEQ) && (symbolB->type == tStr))
    {
      err (true, EF_TYPE_INCOMPATIBLE, symbolB->key);
      erre (EFF_RECOVER);
    }

    mGetToken ();
    symbolC = mParseSimpleExpression();

    // ensure that the types match //
    if (!((((symbolB->type == tInt) || (symbolB->type == tRat)) &&
           ((symbolC->type == tInt) || (symbolC->type == tRat))) ||
          (((symbolB->type == tStr) && (symbolC->type == tStr)))
         )
       )
    {
      err (true, EF_TYPE_INCOMPATIBLE, symbolC->key);
      erre (EFF_RECOVER);
    }

    // Use the string statements if they are string //
    if ((symbolB->type == tStr) && (symbolC->type == tStr))
    {
      if (tac == tacEQ) tac = tacSEQ;
      else if (tac == tacNEQ) tac = tacSNEQ;
    }

    tempgen_check (symbolB);
    tempgen_check (symbolC);
    symbolA = tempgen_make (tRat); // tRat for boolean values

    icm_put (statement_new (tac, symbolA, symbolB, symbolC));
    symbolB = symbolA; // set the output of this loop as the input for next
  }

  return symbolB; // return the simple expression
}

//// mParseSimpleExpression
//// Duties: binary +, -, &, 'or',
Symbol *mParseSimpleExpression ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Symbol *symbolC;
  TAC tac;

  // Parse the first term //
  symbolB = mParseTerm ();

  // Loop to parse subsequent additive operators and terms //
  while ((mCode == codeMINUS) || (mCode == codePLUS) ||
         (mCode == codeOR) || (mCode == codeANDSIGN))
  {
    if (mCode == codeMINUS) tac = tacSUB;
    else if (mCode == codePLUS) tac = tacADD;
    else if (mCode == codeOR) tac = tacOR;
    else if (mCode == codeANDSIGN) tac = tacCAT;

    mGetToken ();
    symbolC = mParseTerm();

    if ((tac == tacSUB) || (tac == tacADD))
    {
      // ensure that types match //
      if (!(((symbolB->type == tInt) || (symbolB->type == tRat)) &&
            ((symbolC->type == tInt) || (symbolC->type == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);

      // This section ensures that if either argument is a rat, //
      //  the resulting value will also be a rat                //
      if ((symbolB->type == tRat) || (symbolC->type == tRat))
      {
        symbolA = tempgen_make (tRat);
      }
      else
      {
        symbolA = tempgen_make (tInt);
      }
    }
    else if (tac == tacOR)
    {
      // ensure that types match //
      if (!( ((symbolB->type == tInt) || (symbolB->type == tRat)) &&
             ((symbolC->type == tInt) || (symbolC->type == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);
      symbolA = tempgen_make (tRat); // rat for booleans
    }
    else if (tac == tacCAT)
    {
      // ensure that types match //
      if (!( (symbolB->type == tStr) && (symbolC->type == tStr)
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);
      symbolA = tempgen_make (tStr);
    }

    icm_put (statement_new (tac, symbolA, symbolB, symbolC));
    symbolB = symbolA; // set the output of this loop as the input for next
  }

  return symbolB; // return the term
}

//// mParseTerm
//// Duties: binary *, /, 'and', ^, 'mod'
Symbol *mParseTerm ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Symbol *symbolC;
  TAC tac;

  // Parse the first factor //
  symbolB = mParseFactor();

  // Loop to parse subsequent multiplicative operators and factors //
  while ((mCode == codeTIMES) || (mCode == codeFORWARDSLASH) ||
         (mCode == codeAND) || (mCode == codeEXPONENT) ||
         (mCode == codeMOD)
        )
  {
    if (mCode == codeTIMES) tac = tacMULT;
    else if (mCode == codeFORWARDSLASH) tac = tacDIV;
    else if (mCode == codeAND) tac = tacAND;
    else if (mCode == codeEXPONENT) tac = tacPOW;
    else if (mCode == codeMOD) tac = tacMOD;

    mGetToken ();
    symbolC = mParseFactor();

    if ((tac == tacMULT) || (tac == tacDIV))
    {
      // ensure that types match //
      if (!(((symbolB->type == tInt) || (symbolB->type == tRat)) &&
            ((symbolC->type == tInt) || (symbolC->type == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);

      // This section ensures that if either argument is a rat, //
      //  the resulting value will also be a rat                //
      if ((symbolB->type == tRat) || (symbolC->type == tRat))
      {
        symbolA = tempgen_make (tRat);
      }
      else
      {
        symbolA = tempgen_make (tInt);
      }
    }
    else if (tac == tacAND)
    {
      // ensure that types match //
      if (!(((symbolB->type == tInt) || (symbolB->type == tRat)) &&
            ((symbolC->type == tInt) || (symbolC->type == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);
      symbolA = tempgen_make (tRat); // rat for boolean
    }
    else if (tac == tacPOW)
    {
      // ensure that types match //
      if (!(((symbolB->type == tInt) || (symbolB->type == tRat)) &&
            (symbolC->type == tInt)
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);
      symbolA = tempgen_make (symbolB->type);
    }
    else if (tac == tacMOD)
    {
      // ensure that types match //
      if (!((symbolB->type == tInt) && (symbolC->type == tInt)))
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      tempgen_check (symbolC);
      symbolA = tempgen_make (symbolB->type);
    }

    icm_put (statement_new (tac, symbolA, symbolB, symbolC));
    symbolB = symbolA;  // set the output of this loop as the input for next
  }

  return symbolB; // return the factor
}

//// mParseFactor ////
//// Duties: identifier, literal, parenthesized expression, unary negative, unary type conversion
Symbol *mParseFactor ()
{
  Symbol *symbolA;
  Symbol *symbolB;
  Type typeA, typeB;

  switch (mCode)
  {
    case codeIdentifier:
    {
      // Search for the identifier //
      symbolA = scope_local_search (mToken->lexeme);
      if (!symbolA)
      {
        // Not found; Undefined identifier error; Enter it in symbol table //
        err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);
        erre (EFF_RECOVER);

        symbolA = scope_local_enter (mToken->lexeme);
      }

      // identifier is 'cinput' //
      if (symbolA == mSymCINPUT)
      {
        symbolA = tempgen_make (tStr);
        icm_put (statement_new (tacCIN, symbolA, NULL, NULL));
      }

      // identifier is a routine call //
      else if (symbolA->type == tRoutine)
      {
        icm_put (statement_new (tacCALL, symbolA, NULL, NULL));

        symbolA = tempgen_make ( ((Routine*)(symbolA->value))->returnType );
        icm_put (statement_new (tacPOP, symbolA, NULL, NULL));
      }

      mGetToken ();
      return symbolA;
      break;
    }
    case codeINT:
    case codeRAT:
    case codeSTR:
    {
      if (mCode == codeINT) typeA = tInt;
      else if (mCode == codeRAT) typeA = tRat;
      else if (mCode == codeSTR) typeA = tStr;

      mGetToken ();
      symbolB = mParseFactor ();
      typeB = symbolB->type;

      // see if the types are the same //
      if (typeA == typeB)
      {
        err (true, EF_UNNECESSARY_TYPE_CONV, gvsTypeStrings[typeA]);
        erre (EFF_RECOVER);

        return symbolB;
      }

      tempgen_check (symbolB);
      symbolA = tempgen_make (typeA);

      if ((typeA == tInt) && (typeB == tRat))
        icm_put (statement_new (tacXRI, symbolA, symbolB, NULL));
      else if ((typeA == tInt) && (typeB == tStr))
        icm_put (statement_new (tacXSI, symbolA, symbolB, NULL));
      else if ((typeA == tRat) && (typeB == tInt))
        icm_put (statement_new (tacXIR, symbolA, symbolB, NULL));
      else if ((typeA == tRat) && (typeB == tStr))
        icm_put (statement_new (tacXSR, symbolA, symbolB, NULL));
      else if ((typeA == tStr) && (typeB == tInt))
        icm_put (statement_new (tacXIS, symbolA, symbolB, NULL));
      else if ((typeA == tStr) && (typeB == tRat))
        icm_put (statement_new (tacXRS, symbolA, symbolB, NULL));

      return symbolA;
      break;
    }
    case codeIntLiteral:
    case codeRatLiteral:
    case codeStrLiteral:
    {
      if (mCode == codeIntLiteral) typeA = tInt;
      else if (mCode == codeRatLiteral) typeA = tRat;
      else if (mCode == codeStrLiteral) typeA = tStr;

      // Search for it and enter it in the symbol table //
      symbolA = scope_global_search (mToken->lexeme);
      if (!symbolA)
      {
        symbolA = scope_global_enter (mToken->lexeme);
        sym_set (symbolA, repLiteral, typeA, true, mToken->value);
      }

      mGetToken ();
      return symbolA;
      break;
    }
    case codeMINUS:
    {
      mGetToken ();
      symbolB = mParseFactor ();

      // ensure that strings are not used //
      if (symbolB->type == tStr)
      {
        err (true, EF_TYPE_INCOMPATIBLE, symbolB->key);
        erre (EFF_RECOVER);
      }

      tempgen_check (symbolB);
      symbolA = tempgen_make (symbolB->type);

      icm_put (statement_new (tacNEG, symbolA, symbolB, NULL));
      return symbolA;
      break;
    }
    case codeLEFTPAREN:
    {
      // Parenthesized subexpression; Call mParseExpression recursively //
      mGetToken ();
      symbolA = mParseExpression ();

      // Check for the closing right parenthesis //
      if (mCode == codeRIGHTPAREN)
      {
        mGetToken ();
      }
      else
      {
        err (true, EF_MISSING_RIGHT_PAREN, 0);
        erre (EFF_IGNORE);
      }

      return symbolA;
      break;
    }
    default:
    {
      err (true, EF_INVALID_EXPRESSION, 0);
      erre (EFF_RECOVER);

      // BUGFIX (29 July 2002, build 0103):
      //   this is added so that a NULL pointer is not returned
      symbolA = scope_global_enter ("<invalid expression>");
      sym_set (symbolA, repLiteral, tStr, true, "<invalid expression>");

      return symbolA;
      break;
    }
  }
}

//////////////
// Special  //
//////////////

void mSkipEOL ()
{
  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == codeEOL)
  {
    mGetToken ();
  }
}

void mEnsureSkipEOL ()
{
  // Make sure that an end of line has been reached //
  mSync (codeEOL, EF_MISSING_EOL);

  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == codeEOL)
  {
    mGetToken ();
  }
}

//// mSync
//// Make sure that one of the codes has been reached
bool mSync (TokenCode aCode, char *aError)
{
  if (mCode != aCode) // code did not match
  {
    if ((strcmp (mToken->lexeme, "") == 0) || (mToken->lexeme == NULL))
      err (true, aError, gvsTokenCodeStrings[mToken->code]);
    else
      err (true, aError, mToken->lexeme);      
    erre (EFF_RECOVER);
    mSeekToEOL ();
    return false;
  }
  return true; // code did match
}

void mSeekToEOL ()
{
  // To allow the parser to continue, skip tokens until an EOL is reached //
  while ((mCode != codeEOL) && (mCode != codeEOF))
  {
    mGetToken ();
  }
}

void mGetToken ()
{
  mToken = lex_get(mToken);
  mCode = mToken->code;
}

#endif


